package com.xmlnamespace.panel.vaadin.ui;

import java.lang.reflect.Method;
import java.util.Collections;
import java.util.Iterator;

import com.vaadin.ui.AbstractComponent;
import com.vaadin.ui.AbstractSingleComponentContainer;
import com.vaadin.ui.Alignment;
import com.vaadin.ui.Component;
import com.vaadin.ui.ComponentContainer;
import com.vaadin.ui.SingleComponentContainer;
import com.vaadin.util.ReflectTools;
import com.xmlnamespace.panel.vaadin.shared.ui.CustomOverlayState;

/**
 * CustomOverlay can be used to add overlays to other components. You can use
 * this to add any component that hovers on top of the component, but it may be
 * easier to use one of the more specific implementations instead:
 * {@link ImageOverlay} or {@link TextOverlay}
 * 
 * This is the server-side part of the component.
 */
public class CustomOverlay extends AbstractComponent implements
        SingleComponentContainer {

    /** The Constant serialVersionUID. */
    private static final long serialVersionUID = 4484572264185406155L;

    /** The Constant COMPONENT_ATTACHED_METHOD. */
    private static final Method COMPONENT_ATTACHED_METHOD = ReflectTools
            .findMethod(ComponentAttachListener.class,
                    "componentAttachedToContainer", ComponentAttachEvent.class);

    /** The Constant COMPONENT_DETACHED_METHOD. */
    private static final Method COMPONENT_DETACHED_METHOD = ReflectTools
            .findMethod(ComponentDetachListener.class,
                    "componentDetachedFromContainer",
                    ComponentDetachEvent.class);

    /**
     * Create empty overlay.
     * 
     * Use {@link #setComponent(Component)} and {@link #setOverlay(Component)}
     * to bind the overlay to a component.
     */
    public CustomOverlay() {
    }

    /**
     * Create empty overlay for a component.
     * 
     * Use {@link #setOverlay(Component)} to add overlay content.
     * 
     * @param overlay
     *            the overlay
     * @param refenceComponent
     *            the refence component
     */
    public CustomOverlay(Component overlay, Component refenceComponent) {
        super();
        setComponent(refenceComponent);
        setOverlay(overlay);
    }

    /**
     * Set the horizontal offset of the overlay from the alignment point.
     * 
     * This is screen coordinates and default is 0.
     * 
     * @param x
     *            Positive for right or negative left offset.
     * @see #setComponentAnchor(Alignment)
     */
    public void setXOffset(int x) {
        getState().x = x;
        markAsDirtyRecursive();
    }

    /**
     * Get the horizontal offset of the overlay from the alignment point.
     * 
     * This is screen coordinates.
     * 
     * @return Positive for right or negative left offset.
     * @see #getComponentAnchor()
     */
    public int getXOffset() {
        return getState().x;
    }

    /**
     * Set the vertical offset of the overlay from the alignment point.
     * 
     * This is screen coordinates and default is 0.
     * 
     * @param y
     *            Positive for downward or negative for upward offset.
     * @see #setComponentAnchor(Alignment)
     */
    public void setYOffset(int y) {
        getState().y = y;
        markAsDirtyRecursive();
    }

    /**
     * Get the vertical offset of the overlay from the alignment point.
     * 
     * This is screen coordinates.
     * 
     * @return Positive for downward or negative upward offset.
     * @see #getComponentAnchor()
     */
    public int getYOffset() {
        return getState().y;
    }

    /**
     * Sets the overlay.
     * 
     * @param overlay
     *            the new overlay
     */
    public void setOverlay(Component overlay) {
        setContent(overlay);
    }

    /**
     * Get the overlay content.
     * 
     * @return the overlay
     */
    public Component getOverlay() {
        return (Component) getState().overlayComp;
    }

    /**
     * Set the reference component.
     * 
     * @param component
     *            The component that this overlay is aligned to.
     */
    public void setComponent(Component component) {
        getState().referenceComp = component;
        markAsDirtyRecursive();
    }

    /**
     * Get the reference component.
     * 
     * @return The component that this overlay is aligned to.
     */
    public Component getComponent() {
        return (Component) getState().referenceComp;
    }

    /**
     * Set the anchor point of the reference component.
     * 
     * The X and Y offsets are relative to this point and overlayAnchor point.
     * 
     * The default is {@link Alignment.TOP_LEFT}
     * 
     * @param anchorPoint
     *            One of the {@link Alignment} constants.
     * @see #setComponent(Component)
     * @see #setXOffset(int)
     * @see #setYOffset(int)
     * @see #setOverlayAnchor(Alignment)
     */
    public void setComponentAnchor(Alignment anchorPoint) {
        Alignment a = anchorPoint != null ? anchorPoint : Alignment.TOP_LEFT;
        getState().align = a.getBitMask();
        markAsDirtyRecursive();
    }

    /**
     * Get the anchor point of the reference component.
     * 
     * The X and Y offsets are relative to this point.
     * 
     * @return align One of the {@link Alignment} constants.
     * @see #setXOffset(int)
     * @see #setYOffset(int)
     * @see #setOverlayAnchor(Alignment)
     */
    public Alignment getComponentAnchor() {
        return new Alignment(getState().align);
    }

    /**
     * Set the alignment point of the overlay component.
     * 
     * The X and Y offsets are relative to this point.
     * 
     * @param overlayAnchorPoint
     *            the new overlay anchor
     * @return align One of the {@link Alignment} constants.
     * @see #setOverlay(Component)
     * @see #setXOffset(int)
     * @see #setYOffset(int)
     * @see #setComponentAnchor(Alignment)
     */
    public void setOverlayAnchor(Alignment overlayAnchorPoint) {
        Alignment a = overlayAnchorPoint != null ? overlayAnchorPoint
                : Alignment.TOP_LEFT;
        getState().overlayAlign = a.getBitMask();
        markAsDirtyRecursive();
    }

    /**
     * Get the alignment point of the overlay component.
     * 
     * The X and Y offsets are relative to this point.
     * 
     * @return align One of the {@link Alignment} constants.
     * @see #setOverlay(Component)
     * @see #setXOffset(int)
     * @see #setYOffset(int)
     * @see #setComponentAnchor(Alignment)
     */
    public Alignment getOverlayAnchor() {
        return new Alignment(getState().overlayAlign);
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.vaadin.ui.AbstractComponent#getState()
     */
    @Override
    protected CustomOverlayState getState() {
        return (CustomOverlayState) super.getState();
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.vaadin.ui.SingleComponentContainer#getComponentCount()
     */
    @Override
    public int getComponentCount() {
        return getOverlay() != null ? 1 : 0;
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.vaadin.ui.HasComponents#iterator()
     */
    @Override
    public Iterator<Component> iterator() {
        if (getOverlay() != null) {
            return Collections.singletonList(getOverlay()).iterator();
        } else {
            return Collections.<Component> emptyList().iterator();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.vaadin.ui.HasComponents.ComponentAttachDetachNotifier#
     * addComponentAttachListener
     * (com.vaadin.ui.HasComponents.ComponentAttachListener)
     */
    @Override
    public void addComponentAttachListener(ComponentAttachListener listener) {
        addListener(ComponentContainer.ComponentAttachEvent.class, listener,
                COMPONENT_ATTACHED_METHOD);
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.vaadin.ui.HasComponents.ComponentAttachDetachNotifier#
     * removeComponentAttachListener
     * (com.vaadin.ui.HasComponents.ComponentAttachListener)
     */
    @Override
    public void removeComponentAttachListener(ComponentAttachListener listener) {
        removeListener(ComponentContainer.ComponentAttachEvent.class, listener,
                COMPONENT_ATTACHED_METHOD);
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.vaadin.ui.HasComponents.ComponentAttachDetachNotifier#
     * addComponentDetachListener
     * (com.vaadin.ui.HasComponents.ComponentDetachListener)
     */
    @Override
    public void addComponentDetachListener(ComponentDetachListener listener) {
        addListener(ComponentContainer.ComponentDetachEvent.class, listener,
                COMPONENT_DETACHED_METHOD);
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.vaadin.ui.HasComponents.ComponentAttachDetachNotifier#
     * removeComponentDetachListener
     * (com.vaadin.ui.HasComponents.ComponentDetachListener)
     */
    @Override
    public void removeComponentDetachListener(ComponentDetachListener listener) {
        removeListener(ComponentContainer.ComponentDetachEvent.class, listener,
                COMPONENT_DETACHED_METHOD);
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.vaadin.ui.SingleComponentContainer#getContent()
     */
    @Override
    public Component getContent() {
        return getOverlay();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * com.vaadin.ui.SingleComponentContainer#setContent(com.vaadin.ui.Component
     * )
     */
    @Override
    public void setContent(Component content) {
        Component oldContent = getContent();
        if (oldContent == content) {
            // do not set the same content twice
            return;
        }
        if (oldContent != null && oldContent.getParent() == this) {
            oldContent.setParent(null);
            fireEvent(new ComponentDetachEvent(this, oldContent));
        }
        getState().overlayComp = content;
        if (content != null) {
            AbstractSingleComponentContainer.removeFromParent(content);

            content.setParent(this);
            fireEvent(new ComponentAttachEvent(this, content));
        }

        markAsDirty();
    }

}
